// Copyright 2024 Proyectos y Sistemas de Mantenimiento SL (eProsima).
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @file doxygen_modules.h
 */

#ifndef FASTCDR_DOXYGEN_MODULES_H_
#define FASTCDR_DOXYGEN_MODULES_H_

/*!
 * @defgroup FASTCDRAPIREFERENCE FastCDR Serializer API Reference
 * @brief FastCDR Serializer internal API groped in modules.
 */

/*!
 * @defgroup EXCEPTIONMODULE Exceptions
 * @ingroup FASTCDRAPIREFERENCE
 * @brief Exceptions used by the FastCDR API.
 * All exceptions defined in this module are thrown by the FastCDR library.
 */

#endif /* FASTCDR_DOXYGEN_MODULES_H_ */
