/*
 * Copyright 2023, Proyectos y Sistemas de Mantenimiento SL (eProsima).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// fold.cpp
#include <iostream>

using namespace std;

template<typename ... Args>
void printer(
        Args&&... args)
{
    (cout << ... << args) << endl;
}

int main()
{
    printer("eProsima", 0b101010, 'x');
}
